#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "instance.h"
#include "client.h"
#include "server.h"
#include "buffer.h"
#include "str-utils.h"
#include "stream.h"

#define FLAG_OUTPUT_READY   (1<<0)
#define FLAG_OUTPUT_CHUNKED (1<<1)

#define SIZE_NAN (~0)

struct aisl_stream
{
  struct buffer          buffer;
  aisl_client_t          client;

  void                 * u_ptr;
  const char           * content_type;

  size_t                 content_length;
  int32_t                content_offset;
  int                    id;
  int                    flags;

  aisl_http_response_t   response;
  aisl_stream_state_t    state;
};


static void
aisl_stream_reset(aisl_stream_t stream, bool initial)
{
  if (!initial)
  {
    aisl_t instance;

    instance = aisl_stream_get_instance(stream);

    aisl_raise(instance, stream, AISL_EVENT_STREAM_CLOSE);
    aisl_unset_callbacks_for(instance, stream);
  }

  buffer_release(&stream->buffer);

  stream->u_ptr          = NULL;
  stream->content_type   = NULL;
  stream->content_length = SIZE_NAN;
  stream->content_offset = 0; /* headers length */
  stream->flags          = 0;
  stream->state          = AISL_STREAM_STATE_IDLE;
  stream->response       = AISL_HTTP_OK;
}


aisl_stream_t
aisl_stream_new(aisl_client_t client, int id)
{
  aisl_stream_t stream = malloc(sizeof(struct aisl_stream));

  if (stream)
  {
    stream->id     = id;
    stream->client = client;
    aisl_stream_reset(stream, true);
  }

  return stream;
}


void
aisl_stream_free(aisl_stream_t stream)
{
  aisl_stream_reset(stream, false);
  free(stream);
}


/* Why it was here?
static int
aisl_stream_write(aisl_stream_t stream, const char * data, uint32_t d_len)
{
  return buffer_add( &stream->buffer, data, d_len);
}


__attribute__ ((visibility ("default") ))
void
aisl_cancel(aisl_stream_t stream)
{
  aisl_client_close( stream->client );
}
*/


__attribute__ ((visibility ("default") ))
void *
aisl_get_context(aisl_stream_t s)
{
  return s->u_ptr;
}


__attribute__ ((visibility ("default") ))
void
aisl_set_context(aisl_stream_t s, void * u_ptr)
{
  s->u_ptr = u_ptr;
}


__attribute__ ((visibility ("default") ))
aisl_client_t
aisl_get_client(aisl_stream_t s)
{
  return s->client;
}


__attribute__ ((visibility ("default") ))
aisl_server_t
aisl_get_server(aisl_stream_t s)
{
  return aisl_client_get_server(s->client);
}


__attribute__ ((visibility ("default") ))
aisl_http_version_t
aisl_get_http_version(aisl_stream_t s)
{
  return aisl_client_get_http_version(s->client);
}


__attribute__ ((visibility ("default") ))
void
aisl_reject(aisl_stream_t s)
{
  aisl_client_close( s->client );
}


static char *
aisl_stream_get_response_begin(aisl_stream_t stream)
{
  char * r;

  r = str_printf(
        "%s %d %s\r\n"
        "Server: AISL\r\n"
        "Connection: %s\r\n\r\n",
        aisl_get_http_version(stream),
        stream->response,
        aisl_http_response_to_string(stream->response),
        aisl_client_is_keepalive(stream->client) ? "keep-alive" : "close"
      );

  return r;
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_response(aisl_stream_t stream, aisl_http_response_t  status_code,
                                    const char           *content_type,
                                    uint32_t              content_length)
{
  char  * pch;
  int     l;


  /* check if those headers were already sent */
  if (stream->state > AISL_STREAM_STATE_READY) return AISL_IDLE;

  stream->response       = status_code;
  stream->content_type   = content_type;
  stream->content_length = content_length;

  if ( !(pch = aisl_stream_get_response_begin(stream)) )
    return AISL_MALLOC_ERROR;

  l = strlen(pch);
  stream->content_offset = l-2;

  buffer_clear(&stream->buffer, content_length);

  l = buffer_add( &stream->buffer, pch, l );
  free(pch);

  if (l == -1) return AISL_MALLOC_ERROR;

  if (content_length)
  {
    if (!aisl_header_printf( stream, "Content-Length", "%u", content_length ))
      return AISL_MALLOC_ERROR;
  }

  if (content_type)
  {
    if (!aisl_header( stream, "Content-Type", content_type ))
      return AISL_MALLOC_ERROR;
  }

  stream->state = AISL_STREAM_STATE_SEND_HEADER;

  return AISL_SUCCESS;
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_flush(aisl_stream_t s)
{
  if ( ! s->content_length )
  {
    s->content_length = s->buffer.size - s->content_offset - 2;
    if (!aisl_header_printf(s, "Content-Length", "%u", s->content_length))
      return AISL_MALLOC_ERROR;
  }

  /*
  fprintf(stdout, "(%lu bytes)------->\n", stream->buffer->size);
  fwrite(stream->buffer->data, 1, stream->buffer->size, stdout);
  fprintf(stdout, "<------\n");
  */
  s->state = AISL_STREAM_STATE_READY;

  s->flags |= FLAG_OUTPUT_READY;

  return AISL_SUCCESS;
}


__attribute__ ((visibility ("default") ))
int
aisl_header(aisl_stream_t stream, const char *key, const char *value)
{
  int ret;
  char * pch;

  if ( (pch = str_printf("%s: %s\r\n", key, value)) != NULL )
  {
    ret = strlen(pch);
    if ( buffer_insert(
           &stream->buffer,
           stream->content_offset,
           pch,
           ret
         ) == -1 )
    {
      ret = -1;
    }
    else
      stream->content_offset += ret;

    free(pch);
  }
  else
    ret = -1;

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_header_printf(aisl_stream_t stream, const char *key,
                                         const char *f_value, ...)
{
  int ret;

  va_list arg;
  va_start(arg, f_value);

  ret = aisl_header_vprintf(stream, key, f_value, arg);

  va_end(arg);

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_header_vprintf(aisl_stream_t stream, const char *key,
                                          const char *format,
                                          va_list     args)
{
  int ret;
  char * value;

  if ( (value = str_vprintf(format, args)) != NULL )
  {
    ret = aisl_header( stream, key, value );
    free(value);
  }
  else
    ret = -1;

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_printf(aisl_stream_t stream, const char *format, ...)
{
  va_list arg;
  va_start(arg, format);

  int result = aisl_vprintf(stream, format, arg);

  va_end(arg);

  /* No need to update length there, because vprintf do that
   * 
   * if (stream->c_length_unknown)
    stream->c_length += result;
    */


  return result;

}


__attribute__ ((visibility ("default") ))
int
aisl_vprintf(aisl_stream_t stream, const char *format, va_list args)
{
  int    result;
  char * r;

  if ( (r = str_vprintf(format, args)) != NULL)
  {
    result = strlen(r);
    if (buffer_add(&stream->buffer, r, result) == -1)
      result = -1;

    free(r);
  }
  else
    result = -1;

  return result;
}


__attribute__ ((visibility ("default") ))
int
aisl_write(aisl_stream_t stream, const char *data, int d_len)
{
  if (d_len < 0)
    d_len = strlen(data);

  if (buffer_add(&stream->buffer, data, d_len) == -1)
    d_len = -1;

  return d_len;
}


__attribute__ ((visibility ("default") ))
int
aisl_puts(const char *str, aisl_stream_t stream)
{
  return aisl_write( stream, str, strlen(str));
}


